from typing import Literal

from fastapi import APIRouter, Depends, HTTPException, Request, Response
from pydantic import BaseModel
from sqlalchemy import or_, orm, select, update
from sqlalchemy.orm import contains_eager, lazyload, sessionmaker

from ultima_scraper_db.databases.ultima_archive.api.client import UAClient
from ultima_scraper_db.databases.ultima_archive.schemas.templates.site import (
    NotificationModel,
)

router = APIRouter(
    prefix="/notifications",
    tags=["notifications"],
    responses={404: {"description": "Not found"}},
)


@router.get("/{site_name}/unsent/{notification_site}")
async def get_notifications(
    request: Request,
    site_name: str,
    notification_site: Literal["discord", "telegram"],
    page: int = 1,
    limit: int = 100,
):
    database_api = UAClient.database_api
    async with database_api.create_site_api(site_name) as site_api:
        notifications = await site_api.get_notifications(
            notification_site, sent=False, page=page, limit=limit
        )
        return notifications


@router.patch("/{site_name}/{notification_site}/{notification_id}")
async def complete_notification(
    request: Request, site_name: str, notification_site: str, notification_id: int
):
    database_api = UAClient.database_api
    async with database_api.create_site_api(site_name) as site_api:
        db_session = site_api.get_session()
        stmt = update(NotificationModel).where(NotificationModel.id == notification_id)
        if notification_site == "telegram":
            stmt = stmt.values(sent_telegram=True)
        else:
            stmt = stmt.values(sent_discord=True)
        await db_session.execute(stmt)
        await db_session.commit()
        return Response(status_code=204)



from typing import Any

import aiohttp
from fastapi import APIRouter, FastAPI
from slowapi import Limiter, _rate_limit_exceeded_handler  # type: ignore
from slowapi.errors import RateLimitExceeded
from slowapi.middleware import SlowAPIMiddleware
from slowapi.util import get_remote_address


class RestAPI(FastAPI):
    Session = aiohttp.ClientSession
    limiter = Limiter(key_func=get_remote_address)

    def __init__(self, **extra: Any):
        super().__init__(**extra)
        self.state.limiter = self.limiter
        self.add_exception_handler(RateLimitExceeded, _rate_limit_exceeded_handler)  # type: ignore
        self.add_middleware(SlowAPIMiddleware)

    def include_routers(self, routers: list[APIRouter]):
        [self.include_router(x) for x in routers]



import re
from dataclasses import asdict, dataclass
from typing import Any, Generic, List, Literal, Optional, TypeVar

import ultima_scraper_api
from fastapi import APIRouter, Depends, HTTPException, Query, Request, Security
from fastapi.encoders import jsonable_encoder
from fastapi.routing import APIRouter
from inflection import singularize, underscore
from pydantic import BaseModel, Field
from pydantic.generics import GenericModel
from sqlalchemy import Case, UnaryExpression, nullslast, or_, orm, select
from sqlalchemy.orm.attributes import InstrumentedAttribute
from sqlalchemy.sql import case, func, or_, select
from typing_extensions import Literal
from ultima_scraper_api.apis.auth_streamliner import datetime

from ultima_scraper_db.databases.ultima_archive.api.client import (
    UAClient,
    get_ua_client,
)
from ultima_scraper_db.databases.ultima_archive.api.routers.users import (
    AdvancedOptions,
    restricted,
)
from ultima_scraper_db.databases.ultima_archive.schemas.templates.site import (
    ContentMediaAssoModel,
    MediaModel,
    UserInfoModel,
)
from ultima_scraper_db.databases.ultima_archive.site_api import (
    MessageModel,
    PostModel,
    UserAliasModel,
    UserModel,
)
from ultima_scraper_db.helpers import extract_identifier_from_url

router = APIRouter(
    prefix="/media",
    tags=["media"],
    responses={404: {"description": "Not found"}},
)


@router.get("/")
async def get_media(
    request: Request,
    site_name: str,
    media_id: int,
    ua_client: UAClient = Depends(get_ua_client),
):
    database_api = ua_client.database_api


class MediaIDs(BaseModel):
    media_ids: List[int]


@router.post("/{site_name}/bulk")
async def get_bulk_media(
    site_name: str,
    item: MediaIDs,
    ua_client: UAClient = Depends(get_ua_client),
):
    database_api = ua_client.database_api
    async with database_api.create_site_api(site_name) as site_api:
        stmt = select(MediaModel).where(MediaModel.id.in_(item.media_ids))
        result = await site_api.get_session().scalars(stmt)
        media_list = result.unique().all()
        return media_list


class ContentIDs(BaseModel):
    content_ids: List[int]


@router.post("/{site_name}/{content_type}/bulk")
async def get_bulk_media_by_content_id(
    site_name: str,
    content_type: str,
    item: ContentIDs,
    ua_client: UAClient = Depends(get_ua_client),
):
    database_api = ua_client.database_api
    async with database_api.create_site_api(site_name) as site_api:
        content_field = f"{singularize(content_type)}_id"
        stmt = (
            select(MediaModel)
            .join(
                ContentMediaAssoModel, ContentMediaAssoModel.media_id == MediaModel.id
            )
            .where(getattr(ContentMediaAssoModel, content_field).in_(item.content_ids))
        )
        result = await site_api.get_session().scalars(stmt)
        media_list = result.unique().all()
        return media_list



from datetime import datetime

from fastapi import APIRouter, Depends, HTTPException, Query, Request, Response
from pydantic import BaseModel
from sqlalchemy import or_, orm, select, update
from sqlalchemy.orm import contains_eager, lazyload, sessionmaker

from ultima_scraper_db.databases.ultima_archive.api.client import UAClient
from ultima_scraper_db.databases.ultima_archive.schemas.templates.site import (
    MediaDetectionModel,
)

router = APIRouter(
    prefix="/media_detections",
    tags=["media_detections"],
    responses={404: {"description": "Not found"}},
)


@router.post("/")
async def get_detected_media(
    filters: list[MediaDetectionModel.MediaDetectionFilter],
    site_name: str,
    user_id: int | None = None,
    page: int = Query(1, alias="page", gt=0),
    limit: int = Query(10, alias="limit", gt=0),
    sex: int | None = None,
    category: str | None = None,
):

    database_api = UAClient.database_api
    async with database_api.create_site_api(site_name) as site_api:
        stmt = MediaDetectionModel().filter_stmt(
            filters, sex, user_id=user_id, category=category
        )
        offset = (page - 1) * limit
        temp = await site_api.get_session().scalars(stmt.offset(offset).limit(limit))
        media_detections = temp.all()
        # [await x.awaitable_attrs.media for x in media_detections]
        return media_detections



import mimetypes
import re
from dataclasses import asdict, dataclass
from pathlib import Path
from typing import Any, Generic, List, Literal, Optional, TypeVar

import aiofiles
import ultima_scraper_api
from fastapi import (
    APIRouter,
    Depends,
    HTTPException,
    Query,
    Request,
    Response,
    Security,
)
from fastapi.encoders import jsonable_encoder
from fastapi.responses import StreamingResponse
from fastapi.routing import APIRouter
from pydantic import BaseModel, Field
from pydantic.generics import GenericModel
from sqlalchemy import Case, UnaryExpression, nullslast, or_, orm, select
from sqlalchemy.orm.attributes import InstrumentedAttribute
from sqlalchemy.sql import case, func, or_, select
from typing_extensions import Literal
from ultima_scraper_api.apis.auth_streamliner import datetime
from ultima_scraper_collection.config import site_config_types
from ultima_scraper_collection.managers.filesystem_manager import FilesystemManager

from ultima_scraper_db.databases.ultima_archive.api.client import (
    UAClient,
    get_ua_client,
)
from ultima_scraper_db.databases.ultima_archive.api.routers.users import (
    AdvancedOptions,
    restricted,
)
from ultima_scraper_db.databases.ultima_archive.schemas.templates.site import (
    MediaModel,
    UserInfoModel,
)
from ultima_scraper_db.databases.ultima_archive.site_api import (
    MessageModel,
    PostModel,
    UserAliasModel,
    UserModel,
)
from ultima_scraper_db.helpers import extract_identifier_from_url

router = APIRouter(
    prefix="/files",
    tags=["files"],
    responses={404: {"description": "Not found"}},
)


async def iterfile(
    file_path: Path, start: int, size: int, chunk_size: int = 1024 * 1024
):
    async with aiofiles.open(file_path, "rb") as f:
        await f.seek(start)
        bytes_left = size
        while bytes_left > 0:
            read_size = min(bytes_left, chunk_size)
            chunk = await f.read(read_size)
            if not chunk:
                break
            yield chunk
            bytes_left -= len(chunk)


@router.api_route("/", methods=["GET", "HEAD"])
async def get_file(
    request: Request,
    site_name: str,
    media_id: int,
    ua_client: UAClient = Depends(get_ua_client),
):
    redis = ua_client.redis
    cache_key = f"filepath:{site_name}:{media_id}"
    cached_path_bytes = await redis.get(cache_key)

    file_path = None
    if cached_path_bytes:
        cached_path = Path(cached_path_bytes.decode())
        if cached_path.exists():
            file_path = cached_path
            # Extend cache validity
            await redis.expire(cache_key, 86400 * 7)  # 24 hours * 7 days
        else:
            # Stale cache, remove it
            await redis.delete(cache_key)

    if not file_path:
        database_api = ua_client.database_api
        datascraper = ua_client.datascraper_manager.select_datascraper(site_name)
        assert datascraper

        fsm = datascraper.filesystem_manager

        async with database_api.create_site_api(site_name) as site_db_api:
            stmt = (
                select(MediaModel)
                .where(MediaModel.id == media_id)
                .options(
                    orm.selectinload(MediaModel.user).selectinload(UserModel.aliases)
                )
            )
            result = await site_db_api.get_session().scalar(stmt)

            if not result:
                raise HTTPException(status_code=404, detail="Media not found")

            username = result.user.username
            aliases = [x.username for x in result.user.aliases]
            usernames = [username] + aliases

            performer_directory = await fsm.performer_directory(
                datascraper.api.site_name, usernames
            )
            if not performer_directory:
                raise HTTPException(
                    status_code=404, detail="Performer directory not found"
                )
            if not result.url:
                raise HTTPException(status_code=404, detail="Media URL not found")
            filename = result.url.split("/")[-1].split("?")[0]

            # NOTE: This recursive glob is very slow and the main cause of timeouts.
            # For a long-term fix, consider storing the file path directly in the database.
            found_files: list[Path] = [
                f for f in performer_directory.rglob("*") if filename in f.name
            ]

            if not found_files:
                raise HTTPException(status_code=404, detail="File not found")

            file_path = found_files[0]
            await redis.set(cache_key, str(file_path.resolve()), ex=86400)

    file_size = file_path.stat().st_size
    mime_type, _ = mimetypes.guess_type(file_path.name)
    mime_type = mime_type or "application/octet-stream"
    # Common headers for both GET and HEAD
    headers = {
        "Content-Disposition": f'inline; filename="{file_path.name}"',
        "Accept-Ranges": "bytes",
        "Content-Length": str(file_size),
    }

    # If it's a HEAD request, return headers immediately without a body.
    if request.method == "HEAD":
        return Response(
            content=None,
            status_code=200,
            headers=headers,
            media_type=mime_type,
        )

    # --- Logic for GET requests below ---

    range_header = request.headers.get("Range")
    if range_header:
        range_match = re.match(r"bytes=(\d+)-(\d*)", range_header)
        if range_match:
            start_byte = int(range_match.group(1))
            end_byte_str = range_match.group(2)
            end_byte = int(end_byte_str) if end_byte_str else file_size - 1

            if start_byte >= file_size:
                raise HTTPException(
                    status_code=416, detail="Request range not satisfiable"
                )
            end_byte = min(end_byte, file_size - 1)
            content_length = end_byte - start_byte + 1

            headers["Content-Length"] = str(content_length)
            headers["Content-Range"] = f"bytes {start_byte}-{end_byte}/{file_size}"

            return StreamingResponse(
                iterfile(file_path, start=start_byte, size=content_length),
                status_code=206,
                headers=headers,
                media_type=mime_type,
            )

    # Full file response for GET without range header
    return StreamingResponse(
        iterfile(file_path, start=0, size=file_size),
        headers=headers,
        media_type=mime_type,
    )


@router.get("/subtitles/{filename}")
async def get_subtitle(
    request: Request, filename: str, ua_client: UAClient = Depends(get_ua_client)
):
    ua_client = get_ua_client(request)
    ua_client.config.settings.subtitles_directory
    subs_dir = ua_client.config.settings.subtitles_directory
    assert subs_dir is not None, "Subtitles directory is not set in config"
    file_path = subs_dir / filename
    file_path = file_path.with_suffix(".srt")
    filename = file_path.name

    if not file_path.exists() or not file_path.is_file():
        raise HTTPException(status_code=404, detail="Subtitle file not found")

    return StreamingResponse(
        iterfile(file_path, start=0, size=file_path.stat().st_size),
        media_type="application/x-subrip",
        headers={"Content-Disposition": f'attachment; filename="{filename}"'},
    )



from typing import Any, Sequence

from fastapi import APIRouter, Depends, HTTPException, Query, Request, Security
from fastapi.encoders import jsonable_encoder
from fastapi.exceptions import HTTPException
from pydantic import BaseModel
from sqlalchemy import (  # Import func from sqlalchemy, not sqlalchemy.orm
    func,
    nullslast,
    or_,
    orm,
    select,
    update,
)
from sqlalchemy.orm import lazyload, selectinload
from sqlalchemy.sql import distinct
from ultima_scraper_api.apis.onlyfans.authenticator import OnlyFansAuthModel

from ultima_scraper_db.databases.ultima_archive.api.client import (
    UAClient,
    get_ua_client,
)
from ultima_scraper_db.databases.ultima_archive.schemas.templates.site import (
    MediaModel,
    MessageModel,
    PostModel,
    UserInfoModel,
    UserModel,
)
from ultima_scraper_db.databases.ultima_archive.site_api import ContentManager

restricted = (
    lazyload(UserModel.user_auths_info),
    orm.defer(UserModel.favorite),
    orm.defer(UserModel.balance),
    orm.defer(UserModel.spend),
    orm.defer(UserModel.updated_at),
    orm.defer(UserModel.created_at),
)


router = APIRouter(
    prefix="/users",
    tags=["users"],
    responses={404: {"description": "Not found"}},
)


class Item(BaseModel):
    name: str
    description: str | None = None
    price: float
    tax: float | None = None


@router.get("/")
async def get_users(
    site_name: str,
    page: int,
    limit: int,
    order_by: str | None = None,
    order_direction: str = "asc",
    ua_client: UAClient = Depends(get_ua_client),
):
    database_api = ua_client.database_api

    # Ensure valid order direction
    if order_direction.lower() not in ["asc", "desc"]:
        raise HTTPException(
            status_code=400, detail="Invalid order_direction, must be 'asc' or 'desc'."
        )

    offset = max(0, (page - 1) * limit)
    async with database_api.create_site_api(site_name) as site_db_api:

        # First, create a subquery that counts paid posts for each user

        # Create subqueries for both paid posts and messages
        paid_posts_subquery = (
            select(PostModel.user_id, func.count().label("posts_ppv_count"))
            .where(PostModel.paid.is_(True))
            .group_by(PostModel.user_id)
            .subquery()
        )

        paid_messages_subquery = (
            select(MessageModel.user_id, func.count().label("messages_ppv_count"))
            .where(MessageModel.paid.is_(True))
            .group_by(MessageModel.user_id)
            .subquery()
        )

        # Subquery for last post date per user
        last_post_date_subquery = (
            select(
                PostModel.user_id,
                func.max(PostModel.created_at).label("last_posted_date"),
            )
            .group_by(PostModel.user_id)
            .subquery()
        )

        # Main query with joins to both subqueries
        stmt = (
            select(
                UserModel,
                (
                    func.coalesce(paid_posts_subquery.c.posts_ppv_count, 0)
                    + func.coalesce(paid_messages_subquery.c.messages_ppv_count, 0)
                ).label("ppv_count"),
                last_post_date_subquery.c.last_posted_date,
            )
            .outerjoin(
                paid_posts_subquery, UserModel.id == paid_posts_subquery.c.user_id
            )
            .outerjoin(
                paid_messages_subquery, UserModel.id == paid_messages_subquery.c.user_id
            )
            .outerjoin(
                last_post_date_subquery,
                UserModel.id == last_post_date_subquery.c.user_id,
            )
            .join(UserInfoModel, UserModel.id == UserInfoModel.user_id)
            .where(UserModel.performer.is_(True))
            .where(
                or_(
                    UserModel.active.is_(True), UserInfoModel.downloaded_at.is_not(None)
                )
            )
            .offset(offset)
            .limit(limit)
            .options(*restricted)
            .options(orm.selectinload(UserModel.user_info))
        )

        # Apply ordering
        if order_by in ["downloaded_at", "size"]:
            field = (
                UserModel.downloaded_at
                if order_by == "downloaded_at"
                else UserInfoModel.size
            )
            if order_direction.lower() == "desc":
                stmt = stmt.order_by(nullslast(field.desc()), UserModel.id)
            else:
                stmt = stmt.order_by(field, UserModel.id)
        else:
            stmt = stmt.order_by(UserModel.id)

        results = await site_db_api.get_session().execute(stmt)

        # Process results
        final_users: list[dict[str, Any]] = []
        for user, ppv_count, last_posted_date in results:
            temp_user = jsonable_encoder(user)
            temp_user["user_info"]["ppv_count"] = ppv_count
            temp_user["last_posted_at"] = last_posted_date
            final_users.append(temp_user)

        return final_users


class AdvancedOptions(BaseModel):
    content: bool = False
    subscribers: bool = False
    subscriptions: bool = False
    auths: bool = False
    bought_contents: bool = False


@router.get("/{site_name}/{identifier}")
async def read(
    site_name: str,
    identifier: str,
    options: AdvancedOptions = Depends(),
    ua_client: UAClient = Depends(get_ua_client),
):
    database_api = ua_client.database_api
    site_api = ua_client.select_site_api(site_name)

    async with database_api.create_site_api(site_name) as site_db_api:
        user = await site_db_api.get_user(
            identifier,
            load_aliases=True,
            load_user_info=True,
            load_remote_urls=True,
            extra_options=restricted,
        )
        if user:
            if options.subscriptions:
                await user.awaitable_attrs.subscriptions
            if options.subscribers:
                await user.awaitable_attrs.subscribers
            if options.auths:
                await user.awaitable_attrs.user_auths_info
            if options.bought_contents:
                await user.awaitable_attrs.bought_contents
            await user.awaitable_attrs.aliases
            await user.awaitable_attrs.user_info
        else:
            authed = await site_api.login(guest=True)
            if authed:
                site_user = await authed.get_user(identifier)
                if site_user:
                    db_user = await site_db_api.get_user(
                        site_user.id,
                        extra_options=restricted,
                    )
                    user = await site_db_api.create_or_update_user(
                        site_user, db_user, performer_optimize=True
                    )
                    user.content_manager = None
                    await user.awaitable_attrs.aliases
                    await user.awaitable_attrs.user_info
                    await user.awaitable_attrs.remote_urls
                await authed.authenticator.close()
        if user:
            user_info = jsonable_encoder(user.user_info, exclude={"user"})
            user = jsonable_encoder(user, exclude={"user_info"})
            user["user_info"] = user_info
            if options:
                if options.content:
                    posts = await site_db_api.get_posts(user["id"], load_media=False)
                    user["posts"] = posts
                    messages = await site_db_api.get_messages(
                        user["id"], load_media=False
                    )
                    user["messages"] = messages
                    pass
        return user


@router.get("/bulk/{site_name}/{identifiers}")
async def read_bulk(
    site_name: str, identifiers: str, ua_client: UAClient = Depends(get_ua_client)
):
    database_api = ua_client.database_api

    async with database_api.create_site_api(site_name) as site_db_api:
        temp_identifiers = identifiers.split(",")
        final_identifiers = [int(x) if x.isdigit() else x for x in temp_identifiers]
        users = await site_db_api.get_users(
            final_identifiers,
            load_aliases=True,
            load_user_info=True,
            load_remote_urls=True,
            extra_options=restricted,
        )
        return users


@router.post("/advanced/{site_name}/{identifier}")
async def read_advanced(
    site_name: str,
    identifier: int | str,
    options: AdvancedOptions | None = None,
    ua_client: UAClient = Depends(get_ua_client),
):
    database_api = ua_client.database_api

    async with database_api.create_site_api(site_name) as site_db_api:
        user = await site_db_api.get_user(identifier, extra_options=restricted)
        if user:
            if options:
                if options.subscriptions:
                    await user.awaitable_attrs.subscriptions
                if options.subscribers:
                    await user.awaitable_attrs.subscribers
                if options.auths:
                    await user.awaitable_attrs.user_auths_info
                if options.bought_contents:
                    await user.awaitable_attrs.bought_contents
            await user.awaitable_attrs.aliases
            await user.awaitable_attrs.user_info
        return user


@router.get("/check/{site_name}/{identifier}")
async def check_user(
    site_name: str, identifier: int | str, ua_client: UAClient = Depends(get_ua_client)
):
    database_api = ua_client.database_api
    site_api = ua_client.select_site_api(site_name)
    has_active_subscription = False
    async with database_api.create_site_api(site_name) as site_db_api:
        user = await site_db_api.get_user(identifier)
        if user:
            await user.awaitable_attrs.subscribers
            db_users = [subscriber.user for subscriber in user.subscribers]
            if db_users:
                db_user = db_users[0]
                for db_auth in db_user.find_authed_buyers(
                    await db_user.find_buyers(
                        active=None if db_user.favorite else True,
                        active_user=True,
                    )
                ):
                    auth_details = db_auth.convert_to_auth_details(site_api.site_name)
                    authed = await site_api.login(auth_details.export())
                    if authed and authed.is_authed():
                        if isinstance(authed, OnlyFansAuthModel):
                            if authed.issues:
                                continue
                        subscription = await authed.get_subscriptions([user.id])
                        if subscription:
                            has_active_subscription = True
                            break
                    else:
                        await db_auth.deactivate()
    return {"has_active_subscription": has_active_subscription}


@router.get("/check/paid_content/{site_name}/{identifier}")
async def check_paid_content(
    site_name: str,
    identifier: int | str,
    only_downloaded: bool = False,
    ua_client: UAClient = Depends(get_ua_client),
):
    database_api = ua_client.database_api
    paid_content: Sequence[MediaModel] = []
    async with database_api.create_site_api(site_name) as site_db_api:
        db_user = await site_db_api.get_user(identifier)
        if db_user:
            db_buyers = await db_user.find_buyers(
                active=None if db_user.favorite else True,
                active_user=True,
            )
            if db_buyers:
                content_manager = ContentManager(db_user)
                paid_content = await content_manager.find_paid_contents(
                    only_downloaded=only_downloaded
                )

    return paid_content


@router.get("/auth/{site_name}/{identifier}")
async def get_auth(
    site_name: str, identifier: int | str, ua_client: UAClient = Depends(get_ua_client)
):
    database_api = ua_client.database_api
    async with database_api.create_site_api(site_name) as db_site:
        user = await db_site.get_user(identifier)
        user_auth_info = None
        if user:
            await user.awaitable_attrs.user_auths_info
            user_auth_info = user.user_auths_info
        return user_auth_info


@router.post("/update/sex/{site_name}/{identifier}/{sex}")
async def update_sex(
    site_name: str,
    identifier: int | str,
    sex: str,
    ua_client: UAClient = Depends(get_ua_client),
):
    database_api = ua_client.database_api
    async with database_api.create_site_api(site_name) as site_db_api:
        user = await site_db_api.get_user(identifier)
        if user:
            await user.awaitable_attrs.user_info
            final_sex = 0 if sex == "female" else 1
            user.user_info.sex = final_sex


@router.get("/optimise_test/{site_name}/{identifier}")
async def optimise_test(
    site_name: str,
    identifier: int | str,
    ua_client: UAClient = Depends(get_ua_client),
):
    database_api = ua_client.database_api
    async with database_api.create_site_api(site_name) as site_db_api:
        db_performer = await site_db_api.get_user(
            identifier, load_content=True, load_media=True
        )
        pass



import re
from typing import TYPE_CHECKING

import sqlalchemy as sa
from sqlalchemy.exc import OperationalError, ProgrammingError
from sqlalchemy.ext.asyncio import AsyncEngine, create_async_engine
from sqlalchemy.orm import Mapped, relationship
from sqlalchemy_utils.functions.database import _set_url_database  # type: ignore
from sqlalchemy_utils.functions.database import _sqlite_file_exists  # type: ignore
from sqlalchemy_utils.functions.database import make_url  # type: ignore
from sqlalchemy_utils.functions.orm import quote  # type: ignore

if TYPE_CHECKING:
    from pathlib import Path

    from ultima_scraper_db.databases.ultima_archive.site_api import MediaManager
TIMESTAMPTZ = sa.TIMESTAMP(timezone=True)


def selectin_relationship(
    *args: str,
    back_populates: str | None = None,
    **kwargs: Mapped[int] | Mapped[str] | Mapped[int | None],
):
    return relationship(*args, lazy="selectin", back_populates=back_populates, **kwargs)


def subquery_relationship(
    *args: str, back_populates: str | None = None, **kwargs: Mapped[int] | Mapped[str]
):
    return relationship(*args, lazy="subquery", back_populates=back_populates, **kwargs)


async def _get_scalar_result(engine: AsyncEngine, sql: sa.TextClause):
    try:
        async with engine.connect() as conn:
            return await conn.scalar(sql)
    except Exception as _e:
        return False


async def test_connection(url: str):
    """
    Test the database connection by executing a simple query.
    Returns True if the connection is successful, False otherwise.
    """
    url_obj = make_url(url)
    engine = create_async_engine(url_obj)

    try:
        async with engine.connect() as conn:
            return True
    except OSError as _e:
        raise Exception("Database connection invalid")


async def database_exists(url: str):
    url_obj = make_url(url)
    database = url_obj.database
    dialect_name = url_obj.get_dialect().name
    engine = None
    try:
        if dialect_name == "postgresql":
            text = "SELECT 1 FROM pg_database WHERE datname='%s'" % database
            for db in (database, "postgres", "template1", "template0", None):
                url_obj = _set_url_database(url_obj, database=db)
                engine = create_async_engine(url_obj)
                try:
                    return bool(await _get_scalar_result(engine, sa.text(text)))
                except (ProgrammingError, OperationalError):
                    pass
            return False

        elif dialect_name == "mysql":
            url_obj = _set_url_database(url_obj, database=None)
            engine = create_async_engine(url_obj)
            text = (
                "SELECT SCHEMA_NAME FROM INFORMATION_SCHEMA.SCHEMATA "
                "WHERE SCHEMA_NAME = '%s'" % database
            )
            return bool(await _get_scalar_result(engine, sa.text(text)))

        elif dialect_name == "sqlite":
            url_obj = _set_url_database(url_obj, database=None)
            engine = create_async_engine(url_obj)
            if database:
                return database == ":memory:" or _sqlite_file_exists(database)
            else:
                # The default SQLAlchemy database is in memory, and :memory: is
                # not required, thus we should support that use case.
                return True
        else:
            text = "SELECT 1"
            try:
                engine = create_async_engine(url_obj)
                return bool(await _get_scalar_result(engine, sa.text(text)))
            except (ProgrammingError, OperationalError):
                return False
    finally:
        if engine:
            await engine.dispose()


async def create_database(
    url: str, encoding: str = "utf8", template: str | None = None
):
    url_obj = make_url(url)
    database = url_obj.database
    dialect_name = url_obj.get_dialect().name
    dialect_driver = url_obj.get_dialect().driver

    if dialect_name == "postgresql":
        url_obj = _set_url_database(url_obj, database="postgres")
    elif dialect_name == "mssql":
        url_obj = _set_url_database(url_obj, database="master")
    elif dialect_name == "cockroachdb":
        url_obj = _set_url_database(url_obj, database="defaultdb")
    elif not dialect_name == "sqlite":
        url_obj = _set_url_database(url_obj, database=None)

    if (dialect_name == "mssql" and dialect_driver in {"pymssql", "pyodbc"}) or (
        dialect_name == "postgresql"
        and dialect_driver in {"asyncpg", "pg8000", "psycopg2", "psycopg2cffi"}
    ):
        engine = create_async_engine(url_obj, isolation_level="AUTOCOMMIT")
    else:
        engine = create_async_engine(url_obj)

    if dialect_name == "postgresql":
        if not template:
            template = "template1"

        async with engine.begin() as conn:
            text = "CREATE DATABASE {} ENCODING '{}' TEMPLATE {}".format(
                quote(conn, database), encoding, quote(conn, template)  # type: ignore
            )
            await conn.execute(sa.text(text))

    elif dialect_name == "mysql":
        async with engine.begin() as conn:
            text = "CREATE DATABASE {} CHARACTER SET = '{}'".format(
                quote(conn, database), encoding  # type: ignore
            )
            await conn.execute(sa.text(text))

    elif dialect_name == "sqlite" and database != ":memory:":
        if database:
            async with engine.begin() as conn:
                await conn.execute(sa.text("CREATE TABLE DB(id int)"))
                await conn.execute(sa.text("DROP TABLE DB"))

    else:
        async with engine.begin() as conn:
            text = f"CREATE DATABASE {quote(conn, database)}"
            await conn.execute(sa.text(text))

    await engine.dispose()


def find_matching_filepaths(
    media_manager: "MediaManager", temp_filepaths: list["Path"]
):
    """
    Find matching file paths for media items without detections.

    Args:
        media_manager: The media manager from the content manager.
        temp_filepaths: List of temporary file paths to search for matches.

    Returns:
        List of found file paths.
    """
    filepaths: list["Path"] = []

    for _key, item in media_manager.medias.items():
        if not item.media_detections:
            found_filepath: "Path" | None = None
            for db_filepath in item.filepaths:
                for temp_filepath in temp_filepaths:
                    if temp_filepath.name in db_filepath.filepath:
                        found_filepath = temp_filepath
                        break
            if found_filepath:
                filepaths.append(found_filepath)

    return filepaths


def has_detected_media(media_manager: "MediaManager"):
    for _key, item in media_manager.medias.items():
        if item.media_detections:
            return True
    return False


def extract_identifier_from_url(url: str) -> str:
    """Extract identifier from the URL."""
    pattern = r"https?://onlyfans\.com/([a-zA-Z0-9_.]+)"
    match = re.match(pattern, url)
    if match:
        return match.group(1)

    return url  # Return the input if it doesn't match the pattern.



"""empty message

Revision ID: 358d71fc1759
Revises: ac840f4e3061
Create Date: 2024-02-13 00:37:33.658319

"""
from alembic import op
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision = '358d71fc1759'
down_revision = 'ac840f4e3061'
branch_labels = None
depends_on = None


def upgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    op.add_column('notifications', sa.Column('authed_user_id', sa.BigInteger(), nullable=True), schema='fansly')
    op.create_foreign_key(None, 'notifications', 'users', ['authed_user_id'], ['id'], source_schema='fansly', referent_schema='fansly')
    op.add_column('notifications', sa.Column('authed_user_id', sa.BigInteger(), nullable=True), schema='onlyfans')
    op.create_foreign_key(None, 'notifications', 'users', ['authed_user_id'], ['id'], source_schema='onlyfans', referent_schema='onlyfans')
    # ### end Alembic commands ###


def downgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_constraint(None, 'notifications', schema='onlyfans', type_='foreignkey')
    op.drop_column('notifications', 'authed_user_id', schema='onlyfans')
    op.drop_constraint(None, 'notifications', schema='fansly', type_='foreignkey')
    op.drop_column('notifications', 'authed_user_id', schema='fansly')
    # ### end Alembic commands ###



import re
from dataclasses import asdict, dataclass
from typing import Any, Generic, List, Literal, Optional, TypeVar

import ultima_scraper_api
from fastapi import APIRouter, Depends, HTTPException, Query, Request, Security
from fastapi.encoders import jsonable_encoder
from fastapi.routing import APIRouter
from pydantic import BaseModel, Field
from pydantic.generics import GenericModel
from sqlalchemy import Case, UnaryExpression, exists, nullslast, or_, orm, select
from sqlalchemy.dialects import postgresql
from sqlalchemy.orm.attributes import InstrumentedAttribute
from sqlalchemy.orm.strategy_options import lazyload
from sqlalchemy.sql import case, func, literal, or_, select, union_all
from sqlalchemy.sql.elements import KeyedColumnElement
from typing_extensions import Literal
from ultima_scraper_api.apis.auth_streamliner import datetime

from ultima_scraper_db.databases.ultima_archive.api.client import (
    UAClient,
    get_ua_client,
)
from ultima_scraper_db.databases.ultima_archive.api.routers.users import (
    AdvancedOptions,
    restricted,
)
from ultima_scraper_db.databases.ultima_archive.schemas.templates.site import (
    ContentMediaAssoModel,
    MediaModel,
    UserInfoModel,
)
from ultima_scraper_db.databases.ultima_archive.site_api import (
    MessageModel,
    PostModel,
    UserAliasModel,
    UserModel,
)
from ultima_scraper_db.helpers import extract_identifier_from_url

T = TypeVar("T")
router = APIRouter(
    prefix="/search",
    tags=["search"],
    responses={404: {"description": "Not found"}},
)


class PaginatedResponse(GenericModel, Generic[T]):
    total: int
    results: List[T]


class Filters:
    def __init__(
        self,
        site_name: str,
        q: str = Query(""),
        order_by: Optional[str] = Query(None),
        order_direction: Literal["asc", "desc"] = Query("asc"),
        has_ppv: Optional[bool] = Query(None),
        prioritize_exact_match: bool = Query(False),
    ):
        self.site_name = site_name
        self.q = q
        self.order_by = order_by
        self.order_direction = order_direction
        self.has_ppv = has_ppv
        self.prioritize_exact_match = prioritize_exact_match

    def as_dict(self, exclude_none: bool = False) -> dict[str, Any]:
        data = self.__dict__
        return (
            {k: v for k, v in data.items() if v is not None} if exclude_none else data
        )


@router.get("/users/{site_name}")
async def search_users(
    filters: Filters = Depends(),
    page: int = Query(1, ge=1),
    limit: int = Query(20, le=50),
    ua_client: UAClient = Depends(get_ua_client),
):
    site_name = filters.site_name
    q = filters.q
    order_by = filters.order_by or "id"
    order_direction = filters.order_direction
    has_ppv = filters.has_ppv

    database_api = ua_client.database_api
    site_api = ua_client.select_site_api(site_name)
    # Ensure valid order direction
    if order_direction.lower() not in ["asc", "desc"]:
        raise HTTPException(
            status_code=400, detail="Invalid order_direction, must be 'asc' or 'desc'."
        )
    offset = (page - 1) * limit  # calculate offset for pagination
    q = q.replace(" ", "")
    q = q.replace("@", "")
    q = extract_identifier_from_url(q)
    async with database_api.create_site_api(site_name) as site_db_api:

        # Create subqueries for both paid posts and messages
        paid_posts_subquery = (
            select(PostModel.user_id, func.count().label("posts_ppv_count"))
            .where(PostModel.paid.is_(True))
            .group_by(PostModel.user_id)
            .subquery()
        )

        paid_messages_subquery = (
            select(MessageModel.user_id, func.count().label("messages_ppv_count"))
            .where(MessageModel.paid.is_(True))
            .group_by(MessageModel.user_id)
            .subquery()
        )
        last_post_date_subquery = (
            select(
                PostModel.user_id,
                func.max(PostModel.created_at).label("last_posted_date"),
            )
            .group_by(PostModel.user_id)
            .subquery()
        )

        stmt = (
            select(
                UserModel,
                (
                    func.coalesce(paid_posts_subquery.c.posts_ppv_count, 0)
                    + func.coalesce(paid_messages_subquery.c.messages_ppv_count, 0)
                ).label("ppv_count"),
                last_post_date_subquery.c.last_posted_date,
            )
            .outerjoin(
                paid_posts_subquery, UserModel.id == paid_posts_subquery.c.user_id
            )
            .outerjoin(
                paid_messages_subquery, UserModel.id == paid_messages_subquery.c.user_id
            )
            .outerjoin(UserAliasModel)
            .outerjoin(
                last_post_date_subquery,
                UserModel.id == last_post_date_subquery.c.user_id,
            )
            .where(
                or_(
                    UserModel.username.ilike(f"%{q}%"),
                    UserAliasModel.username.ilike(f"%{q}%"),
                )
            )
        )
        if has_ppv:
            # Only include users who have at least one paid post or paid message
            stmt = stmt.where(
                or_(
                    paid_posts_subquery.c.posts_ppv_count.isnot(None),
                    paid_messages_subquery.c.messages_ppv_count.isnot(None),
                )
            )

        # Always order by exact match first
        ordering: list[
            Case[Any]
            | UnaryExpression[datetime | int]
            | InstrumentedAttribute[datetime]
            | InstrumentedAttribute[int]
        ] = []

        if filters.prioritize_exact_match:
            ordering.append(
                case((func.lower(UserModel.username) == q.lower(), 0), else_=1)
            )

        # Then add field-based ordering if specified
        if order_by in ["downloaded_at", "size"]:
            field = (
                UserModel.downloaded_at
                if order_by == "downloaded_at"
                else UserInfoModel.size
            )
            if order_direction.lower() == "desc":
                ordering.append(nullslast(field.desc()))
            else:
                ordering.append(field)

        # Always order by id last
        ordering.append(UserModel.id)

        # Apply to query
        stmt = stmt.order_by(*ordering)

        total_stmt = select(func.count(func.distinct(stmt.subquery().c.id)))
        total = await site_db_api.get_session().scalar(total_stmt)
        stmt = (
            stmt.offset(offset)
            .limit(limit)
            .options(*restricted)
            .options(orm.selectinload(UserModel.user_info))
        )
        results = await site_db_api.get_session().execute(stmt)
        final_users: list[dict[str, Any]] = []
        accurate_username = False
        for user, ppv_count, last_posted_date in results.unique():
            print(
                f"Processing user: {user.username}, PPV Count: {ppv_count}, Last Post Date: {last_posted_date}"
            )
            temp_user = jsonable_encoder(user)
            if temp_user["user_info"]:
                temp_user["user_info"]["ppv_count"] = ppv_count
            temp_user["last_posted_at"] = last_posted_date
            if q.lower() == str(user.id) or q.lower() == user.username.lower():
                accurate_username = True
            final_users.append(temp_user)

        _all_are_performers = all(entry.get("performer") for entry in final_users)
        if not final_users or not accurate_username:
            async with site_api.login_context(guest=True) as authed:
                if authed:
                    site_user = await authed.get_user(q)
                    if site_user:
                        db_user = await site_db_api.get_user(
                            site_user.id,
                            extra_options=restricted,
                        )
                        user = await site_db_api.create_or_update_user(
                            site_user, db_user, performer_optimize=True
                        )
                        if site_user.username != q:
                            _user_alias = await site_db_api.create_or_update_user_alias(
                                user, q
                            )
                        user.content_manager = None
                        await user.awaitable_attrs.aliases
                        await user.awaitable_attrs.user_info
                        await user.awaitable_attrs.remote_urls
                        results = await site_db_api.get_session().scalars(stmt)
                        # fix this, we get json encode max recursion error, why? I don't know. But it could be because of the awaitable_attrs
                        # _abc = results.all()
                        final_users = [
                            jsonable_encoder(
                                user, exclude={"user_info": "user", "aliases": "user"}
                            )
                            for user in results.all()
                        ]

        return PaginatedResponse[dict[str, Any]](
            total=total, results=jsonable_encoder(final_users)
        )


@router.get(
    "/{site_name}/content",
    response_model=PaginatedResponse[dict[str, Any]],
)
async def search_content(
    category: Optional[Literal["posts", "messages"]] = None,
    filters: Filters = Depends(),
    user_id: int | None = None,
    has_media: Optional[bool] = Query(None),
    page: int = Query(1, ge=1),
    limit: int = Query(100, le=1000),
    ua_client: UAClient = Depends(get_ua_client),
):

    site_name = filters.site_name
    q = filters.q
    order_by = filters.order_by or "id"
    order_direction = filters.order_direction
    has_ppv = filters.has_ppv

    database_api = ua_client.database_api
    offset = (page - 1) * limit

    async with database_api.create_site_api(site_name) as site_db_api:
        session = site_db_api.get_session()

        if category:
            if category == "posts":
                content_model = PostModel
            elif category == "messages":
                content_model = MessageModel
            else:
                # This path should ideally not be taken due to Literal type hint
                raise HTTPException(status_code=400, detail="Invalid content type")

            # Build base conditions that will be reused for both queries
            base_conditions: list[Any] = []

            if user_id:
                base_conditions.append(content_model.user_id == user_id)
            if q:
                base_conditions.append(content_model.text.ilike(f"%{q}%"))
            if has_ppv is not None:
                base_conditions.append(
                    content_model.price > 0 if has_ppv else content_model.price == 0
                )

            # Handle media filter efficiently
            if has_media is not None:
                assoc = ContentMediaAssoModel.__table__
                if category == "posts":
                    content_fk = assoc.c.post_id
                elif category == "messages":
                    content_fk = assoc.c.message_id

                # Create a more efficient media check
                media_with_url_ids = select(MediaModel.id).where(
                    MediaModel.url.is_not(None)
                )

                media_exists = exists().where(
                    content_fk == content_model.id,
                    assoc.c.media_id.in_(media_with_url_ids),
                )

                if has_media:
                    base_conditions.append(media_exists)
                else:
                    base_conditions.append(~media_exists)

            # COUNT query - use base conditions directly (much faster)
            total_stmt = select(func.count(content_model.id)).where(*base_conditions)
            total = await session.scalar(total_stmt)

            # Main query - start with base conditions
            stmt = select(content_model).where(*base_conditions)
            # Always order by exact match first
            ordering: list[
                Case[Any]
                | UnaryExpression[datetime | int]
                | InstrumentedAttribute[datetime]
                | InstrumentedAttribute[int]
                | KeyedColumnElement[Any]
            ] = []

            if filters.prioritize_exact_match:
                ordering.append(
                    case((func.lower(content_model.text) == q.lower(), 0), else_=1)
                )

            # Field-based ordering
            if order_by in ["id", "user_id"]:
                field = content_model.id if order_by == "id" else content_model.user_id
                if order_direction.lower() == "desc":
                    ordering.append(nullslast(field.desc()))
                else:
                    ordering.append(field)

            # Always order by id last
            ordering.append(content_model.id)

            # Apply ordering and pagination
            stmt = stmt.order_by(*ordering).offset(offset).limit(limit)

            # Don't load media relationships for performance
            stmt = stmt.options(orm.selectinload(content_model.user)).options(
                orm.noload(content_model.media)
            )

            results = await session.scalars(stmt)
            content_items = results.unique().all()

            return PaginatedResponse[dict[str, Any]](
                total=total, results=jsonable_encoder(content_items)
            )
        else:
            # Category is None, search both posts and messages
            def get_conditions(model: type[PostModel] | type[MessageModel]):
                conditions: list[Any] = []
                if user_id:
                    conditions.append(model.user_id == user_id)
                if q:
                    conditions.append(model.text.ilike(f"%{q}%"))
                if has_ppv is not None:
                    conditions.append(model.price > 0 if has_ppv else model.price == 0)
                if has_media is not None:
                    assoc = ContentMediaAssoModel.__table__
                    content_fk = None
                    if model == PostModel:
                        content_fk = assoc.c.post_id
                    elif model == MessageModel:
                        content_fk = assoc.c.message_id

                    if content_fk is not None:
                        media_with_url_ids = select(MediaModel.id).where(
                            MediaModel.url.is_not(None)
                        )
                        media_exists = exists().where(
                            content_fk == model.id,
                            assoc.c.media_id.in_(media_with_url_ids),
                        )
                        if has_media:
                            conditions.append(media_exists)
                        else:
                            conditions.append(~media_exists)
                return conditions

            post_conditions = get_conditions(PostModel)
            message_conditions = get_conditions(MessageModel)

            post_total_stmt = select(func.count(PostModel.id)).where(*post_conditions)
            message_total_stmt = select(func.count(MessageModel.id)).where(
                *message_conditions
            )
            post_total = await session.scalar(post_total_stmt)
            message_total = await session.scalar(message_total_stmt)
            total = (post_total or 0) + (message_total or 0)

            post_stmt = select(
                PostModel.id,
                PostModel.text,
                PostModel.price,
                PostModel.user_id,
                literal("post").label("type"),
            ).where(*post_conditions)
            message_stmt = select(
                MessageModel.id,
                MessageModel.text,
                MessageModel.price,
                MessageModel.user_id,
                literal("message").label("type"),
            ).where(*message_conditions)

            union_stmt = union_all(post_stmt, message_stmt).alias("content")

            ordering: list[
                Case[Any]
                | UnaryExpression[datetime | int]
                | InstrumentedAttribute[datetime]
                | InstrumentedAttribute[int]
                | KeyedColumnElement[Any]
            ] = []
            if filters.prioritize_exact_match and q:
                ordering.append(
                    case((func.lower(union_stmt.c.text) == q.lower(), 0), else_=1)
                )

            if order_by in ["id", "user_id"]:
                field = union_stmt.c.id if order_by == "id" else union_stmt.c.user_id
                if order_direction.lower() == "desc":
                    ordering.append(nullslast(field.desc()))
                else:
                    ordering.append(field)

            ordering.append(union_stmt.c.id)

            stmt = select(union_stmt).order_by(*ordering).offset(offset).limit(limit)
            results = await session.execute(stmt)
            content_items = [dict(row) for row in results.mappings()]

            # Fetch and attach user objects
            user_ids = {item["user_id"] for item in content_items}
            if user_ids:
                users_stmt = select(UserModel).where(UserModel.id.in_(user_ids))
                users_result = await session.scalars(users_stmt)
                users_map = {
                    user.id: jsonable_encoder(user) for user in users_result.unique()
                }
                for item in content_items:
                    item["user"] = users_map.get(item["user_id"])

            return PaginatedResponse[dict[str, Any]](
                total=total, results=jsonable_encoder(content_items)
            )



from fastapi import APIRouter, Depends, HTTPException, Request, Response
from pydantic import BaseModel
from sqlalchemy import or_, orm, select, update
from sqlalchemy.orm import contains_eager, lazyload, sessionmaker

from ultima_scraper_db.databases.ultima_archive.api.client import UAClient
from ultima_scraper_db.databases.ultima_archive.schemas.templates.site import (
    FilePathModel,
    JobModel,
)

router = APIRouter(
    prefix="/client",
    tags=["client"],
    responses={404: {"description": "Not found"}},
)


@router.get("/whoami/{identifier}")
async def whoami(request: Request, identifier: int | str):
    database_api = UAClient.database_api
    async with database_api.create_management_api() as management_api:
        if isinstance(identifier, int):
            server = await management_api.get_server(
                server_id=identifier, server_name=None
            )
        else:
            if identifier.isdigit():
                server = await management_api.get_server(
                    server_id=int(identifier), server_name=None
                )
            else:
                server = await management_api.get_server(
                    server_id=None, server_name=identifier
                )
    return server


@router.post("/")
async def get_ip(request: Request):
    assert request.client
    client_host = request.client.host
    return {"client_ip": client_host}



from typing import Any

from fastapi import APIRouter
from pydantic import BaseModel
from sqlalchemy import ScalarResult, Select, func, or_, select, update
from ultima_scraper_db.databases.ultima_archive.api.client import UAClient
from ultima_scraper_db.databases.ultima_archive.schemas.templates.site import JobModel

restricted = (
    # lazyload(UserModel.user_auth_info),
    # orm.defer(UserModel.performer),
    # orm.defer(UserModel.favorite),
    # orm.defer(UserModel.balance),
    # orm.defer(UserModel.spend),
    # orm.defer(UserModel.updated_at),
    # orm.defer(UserModel.created_at),
)

router = APIRouter(
    prefix="/jobs",
    tags=["jobs"],
    responses={404: {"description": "Not found"}},
)


class JobData(BaseModel):
    class OptionData(BaseModel):
        class MandatoryJob(BaseModel):
            job_name: str
            success_queue_name: str | None = None
            failure_queue_name: str | None = None

        mandatory_jobs: list[MandatoryJob] = []
        ppv_only: bool | None = None
        extra_data: dict[str, Any] = {}

        def create_mandatory_job(
            self,
            job_name: str,
            success_queue_name: str | None = None,
            failure_queue_name: str | None = None,
        ) -> "JobData.OptionData.MandatoryJob":
            # Check if a job with the same job_name already exists
            for job in self.mandatory_jobs:
                if job.job_name == job_name:
                    return job  # Return the existing job if found

            # If no match is found, create and add the new job
            mandatory_job = self.MandatoryJob(
                job_name=job_name,
                success_queue_name=success_queue_name,
                failure_queue_name=failure_queue_name,
            )
            self.mandatory_jobs.append(mandatory_job)
            return mandatory_job

    site_name: str
    user_id: int
    username: str
    category: str
    options: OptionData = OptionData()
    server_id: int = 1
    priority: int = 0
    host_id: int | None = None
    skippable: bool = False
    active: bool | None = None


class UpdateJob(BaseModel):
    id: int | None = None
    active: bool | None = None


@router.post("/")
async def get_jobs(
    job_type: JobData,
    site_name: str,
    page: int = 1,
    limit: int = 100,
):
    database_api = UAClient.database_api

    site_api = database_api.get_site_api(site_name)
    async with site_api as site_api:
        # limit = 100 if limit > 100 else limit
        jobs = await site_api.get_jobs(
            server_id=job_type.server_id,
            performer_id=job_type.performer_id,
            category=job_type.category,
            page=page,
            limit=limit,
            active=job_type.active,
        )
    return jobs


@router.post("/create")
async def create_job(
    job: JobData,
):
    database_api = UAClient.database_api

    site_api = database_api.get_site_api(job.site_name)
    async with site_api as site_api:
        user = await site_api.get_user(job.user_id)
        if user:
            assert job.server_id
            assert job.category
            _db_job = await site_api.create_or_update_job(
                user,
                job.category,
                server_id=job.server_id,
                host_id=job.host_id,
                skippable=job.skippable,
                options=job.options.model_dump() if job.options else None,
                priority=job.priority,
            )
            await user.awaitable_attrs.subscribers
            return user


from datetime import datetime


@router.post("/complete")
async def complete_job(
    job_type: UpdateJob,
    site_name: str,
):
    database_api = UAClient.database_api

    site_api = database_api.get_site_api(site_name)
    async with site_api as site_api:
        stmt = (
            update(JobModel)
            .where(JobModel.id == job_type.id)
            .values(active=job_type.active, completed_at=datetime.now())
        )
        await site_api.get_session().execute(stmt)
        return True


from fastapi import Query


@router.post("/test")
async def test(site_name: str, filepath_str: str = Query(alias="filepath")):
    from pathlib import Path

    from ultima_scraper_collection.config import site_config_types
    from ultima_scraper_collection.managers.filesystem_manager import FilesystemManager

    database_api = UAClient.database_api

    config = UAClient.config
    site_config: site_config_types = config.get_site_config(site_name=site_name)
    fsm = FilesystemManager()
    fsm.activate_directory_manager(site_config)
    filepath = Path(filepath_str).as_posix()
    if not Path(filepath).exists():
        download_directories = site_config.download_setup.directories
        edited_filepath = filepath.split(site_name)[1]
        for directory in download_directories:
            if directory.path:
                temp_path = Path(
                    directory.path, site_name, edited_filepath.removeprefix("/")
                )
                if temp_path.exists():
                    filepath = temp_path
                    return filepath

    site_db_api = database_api.get_site_api(site_name)

    async with site_db_api as site_db_api:
        db_filepaths = await site_db_api.get_filepaths(Path(filepath).name)
        performer_identifiers: list[int | str] = []

        for db_filepath in db_filepaths:
            db_content = await db_filepath.get_content()
            await db_content.awaitable_attrs.user
            db_user = db_content.user
            await db_user.awaitable_attrs.aliases
            performer_identifiers.extend([db_user.id, db_user.username])

        found_path = None

        for performer_identity in performer_identifiers:
            appearances = filepath.count(str(performer_identity))
            if appearances == 1:
                last_position = filepath.rfind(str(performer_identity))
                part1 = Path(filepath[:last_position])
                part2 = Path(filepath[last_position:])

                for directory in site_config.download_setup.directories:
                    if directory.path:
                        temp_unique_path = Path(directory.path, part2)

                        if not temp_unique_path.exists():
                            reversed_parts = []
                            for part in reversed(part1.parts):
                                reversed_parts = [part] + reversed_parts
                                temp_unique_path = Path(
                                    directory.path, Path(*reversed_parts), part2
                                )

                                if temp_unique_path.exists():
                                    found_path = temp_unique_path
                                    break

                if found_path:
                    break
            else:
                if appearances > 1:
                    breakpoint()

        if found_path:
            return found_path
        else:
            performer_usernames = [
                db_user.username,
                *[x.username for x in db_user.aliases],
            ]
            found_username = None
            for performer_username in performer_usernames:
                appearances = filepath.count(performer_username)
                if appearances == 1:
                    found_username = performer_username

            if found_username:
                edited_filepath = filepath.split(site_name)[1]
                performer_usernames.remove(found_username)
                for directory in site_config.download_setup.directories:
                    for performer_username in performer_usernames:
                        edited_filepath_2 = edited_filepath.replace(
                            found_username[0].capitalize(),
                            performer_username[0].capitalize(),
                        )
                        edited_filepath_2 = edited_filepath_2.replace(
                            found_username, performer_username
                        )
                        temp_path = Path(
                            directory.path,
                            site_name,
                            edited_filepath_2.removeprefix("/"),
                        )
                        if temp_path.exists():
                            filepath = temp_path
                            return filepath
    return
